/* ------------------------------------------------------------------------ */
/* xlauto.c - C-Funktionen zur Automatisierung von Microsoft Excel          */
/*                                                                          */
/*                  Copyright Tilman Kuepper 2017.                          */
/*    Distributed under the Boost Software License, Version 1.0.            */
/*        (See accompanying file LICENSE_1_0.txt or copy at                 */
/*               http://www.boost.org/LICENSE_1_0.txt)                      */
/*                                                                          */
/* Letzte Bearbeitung: 11. Maerz 2017                                       */
/* ------------------------------------------------------------------------ */

#define _CRT_SECURE_NO_WARNINGS

#include <windows.h>
#include <oleauto.h>  /* VARIANT-Typ fuer lcc-Compiler */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "xlauto.h"

/* Visual Studio 2010: notwendige Bibliotheken einbinden */
#ifdef _MSC_VER
    #pragma comment(lib, "ole32.lib")
    #pragma comment(lib, "oleaut32.lib")
#endif

/* Bibliotheken fuer lcc-Compiler */
#ifdef __LCC__
    #pragma lib <ole32.lib>
    #pragma lib <oleaut32.lib>
    #pragma lib <uuid.lib>
#endif

/* Funktion "cmd" aufrufen, HRESULT-Rueckgabewert mittels assert pruefen */
#undef HM_CHECK
#define HM_CHECK(cmd) { HRESULT hr_ = (cmd); if(FAILED(hr_)) assert(!#cmd); }

/* Ist die Verbindung zu Microsoft Excel bereits hergestellt? */
static int xl_running = 0;

/* Dispatch-ID zum Zugriff auf Microsoft Excel */
static VARIANT xl_app = { 0 };

/* Anzahl der Elemente in einem Feld ermitteln */
#undef ARR_SIZE
#define ARR_SIZE(arr) (sizeof((arr)) / sizeof((arr)[0]))

/* Visual Studio 2010: Warnung bei Benutzung von C++/CLI vermeiden */
#pragma managed(push, off)

/* ------------------------------------------------------------------------ */
/* Aufruf einer OLE-Methode bzw. Zugriff auf OLE-Eigenschaft. Die Funktion  */
/* stammt aus einer OLE-Beispielapplikation im Microsoft Developer Network  */
/* (B2CSE.exe); sie wurde lediglich leicht angepasst.                       */
/* ------------------------------------------------------------------------ */
static HRESULT AutoWrap(WORD autoType, VARIANT *pvResult, VARIANT *disp, LPOLESTR ptName, INT cArgs, ...)
{
    IDispatch *pDisp = disp->pdispVal;
    DISPID dispID, dispidNamed = DISPID_PROPERTYPUT;
    DISPPARAMS dp = { NULL, NULL, 0, 0 };
    VARIANT args[10]; HRESULT hr; INT i;

    /* Begin variable-argument list */
    va_list marker;
    va_start(marker, cArgs);

    assert(disp && disp->vt == VT_DISPATCH && disp->pdispVal);
    if(!disp || disp->vt != VT_DISPATCH || !disp->pdispVal)
    {
        const char *msg = "Invalid IDispatch passed to AutoWrap()";
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
        _exit(0);  /* ----------------------------- */
    }
    
    /* Get DISPID for name passed */
#if defined(__cplusplus) && !defined(CINTERFACE)
    hr = pDisp->GetIDsOfNames(IID_NULL, &ptName, 1, LOCALE_USER_DEFAULT, &dispID);
#else
    hr = pDisp->lpVtbl->GetIDsOfNames(pDisp, &IID_NULL, &ptName, 1, LOCALE_USER_DEFAULT, &dispID);
#endif

    if(FAILED(hr))
    {
        char msg[1024], name[512];
        wcstombs(name, ptName, ARR_SIZE(name)); name[ARR_SIZE(name) - 1] = 0;
        sprintf(msg, "GetIDsOfNames(\"%s\") failed, err = %08lx", name, hr);
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
        _exit(0);  /* ----------------------------- */
    }
    
    /* Extract arguments */
    assert(cArgs < 10);
    for(i = 0; i < cArgs; i++)
        args[i] = va_arg(marker, VARIANT);
    
    /* Build DISPPARAMS, special case for property-puts */
    dp.cArgs = cArgs, dp.rgvarg = args;
    if(autoType & DISPATCH_PROPERTYPUT)
        dp.cNamedArgs = 1, dp.rgdispidNamedArgs = &dispidNamed;

    /* Call OLE method */
#if defined(__cplusplus) && !defined(CINTERFACE)
    hr = pDisp->Invoke(dispID, IID_NULL, LOCALE_SYSTEM_DEFAULT, autoType, &dp, pvResult, NULL, NULL);
#else
    hr = pDisp->lpVtbl->Invoke(pDisp, dispID, &IID_NULL, LOCALE_SYSTEM_DEFAULT, autoType, &dp, pvResult, NULL, NULL);
#endif

    /*  Falls der Aufruf der OLE-Methode misslungen sein sollte, gibt AutoWrap
     *  den Fehlercode als Rueckgabewert an die aufrufende Funktion zurueck.
     *  Dort wird eine Fehlermeldung ausgegeben - und nicht etwa schon hier...
     */

    /*
    if(FAILED(hr))
    {
        char msg[1024], name[512];
        wcstombs(name, ptName, ARR_SIZE(name)); name[ARR_SIZE(name) - 1] = 0;
        sprintf(msg, "OLE::Invoke(\"%s\") failed, err = %08lx", name, hr);
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
    }
    */

    /* End variable-argument section */
    va_end(marker);
    return hr;
}

/* Visual Studio 2010: Warnung bei Benutzung von C++/CLI vermeiden */
#pragma managed(pop)

/* ------------------------------------------------------------------------ */
/* Verbindung zu Microsoft Excel herstellen. Der Parameter show_window gibt */
/* an, ob Excel sichtbar (XL_SHOW_WINDOW) oder unsichtbar (XL_HIDE_WINDOW)  */
/* gestartet werden soll. Mit dem Parameter filename wird der Name der zu   */
/* oeffnenden Excel-Datei uebergeben, bzw. filename == NULL, falls keine    */
/* Datei geoeffnet werden soll. Rueckgabewert: XL_OK, XL_ERR_NOT_FOUND,     */
/* XL_ERR_NOT_INSTALLED oder XL_ERR_ALREADY_RUNNING                         */
/* ------------------------------------------------------------------------ */
int XL_Connect(const char *filename, int show_window)
{
    DWORD com_server_type = CLSCTX_LOCAL_SERVER | CLSCTX_INPROC_SERVER;
    VARIANT xl_workbooks = { 0 }, arg = { 0 };
    CLSID clsid; HRESULT hr;

    assert(!xl_running);
    if(xl_running) return XL_ERR_ALREADY_RUNNING;

    /* Initialize the OLE Library */
    HM_CHECK(OleInitialize(NULL));

    /* Set Application = CreateObject("Excel.Application") */
    hr = CLSIDFromProgID(L"Excel.Application", &clsid);
    if(FAILED(hr))
    {
        const char *msg = "Failed to start Microsoft Excel - not installed?";
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
        OleUninitialize();
        return XL_ERR_NOT_INSTALLED;
    }

#if defined(__cplusplus) && !defined(CINTERFACE)
    hr = CoCreateInstance(clsid, NULL, com_server_type, IID_IDispatch, (void**)&xl_app.pdispVal);
#else
    hr = CoCreateInstance(&clsid, NULL, com_server_type, &IID_IDispatch, (void**)&xl_app.pdispVal);
#endif

    if(FAILED(hr))
    {
        const char *msg = "Failed to start Microsoft Excel - not installed?";
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
        OleUninitialize();
        return XL_ERR_NOT_INSTALLED;
    }

    xl_app.vt = VT_DISPATCH;
    xl_running = 1;

    /* Application.Visible = True */
    arg.vt = VT_I4, arg.lVal = (show_window == XL_SHOW_WINDOW ? 1 : 0);
    HM_CHECK(AutoWrap(DISPATCH_PROPERTYPUT, NULL, &xl_app, L"Visible", 1, arg));
    HM_CHECK(VariantClear(&arg));

    if(filename != NULL)
    {
        wchar_t wname[1024];
        mbstowcs(wname, filename, ARR_SIZE(wname)); wname[ARR_SIZE(wname) - 1] = 0;
        arg.vt = VT_BSTR; arg.bstrVal = SysAllocString(wname); assert(arg.bstrVal);

        /* Application.Workbooks.Open(filename) */
        HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_workbooks, &xl_app, L"Workbooks", 0));
        hr = AutoWrap(DISPATCH_METHOD, NULL, &xl_workbooks, L"Open", 1, arg);
        HM_CHECK(VariantClear(&arg)); HM_CHECK(VariantClear(&xl_workbooks));

        if(SUCCEEDED(hr))
            return XL_OK;
        else
        {
            /*  Der angegebene Dateiname wurde nicht gefunden. Das Programm
             *  wird dennoch nicht gleich beendet - stattdessen wird eine
             *  Fehlermeldung auf dem Bildschirm ausgegeben und ein leeres
             *  Arbeitsblatt geladen (Application.Workbooks.Add, s. u.).
             */
            char msg[1024];
            sprintf(msg, "File not found: \"%s\"", filename);
            MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
        }
    }

    /* Leeres Arbeitsblatt laden: Application.Workbooks.Add  */
    HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_workbooks, &xl_app, L"Workbooks", 0));
    HM_CHECK(AutoWrap(DISPATCH_METHOD, NULL, &xl_workbooks, L"Add", 0));
    HM_CHECK(VariantClear(&xl_workbooks));
    return filename == NULL ? XL_OK : XL_ERR_NOT_FOUND;
}

/* ------------------------------------------------------------------------ */
/* Verbindung zu Microsoft Excel beenden, bei mode == XL_STAY_OPEN bleibt   */
/* Excel weiterhin offen, bei mode == XL_CLOSE wird Excel geschlossen.      */
/* Rueckgabewert: XL_OK oder XL_ERR_NOT_RUNNING                             */
/* ------------------------------------------------------------------------ */
int XL_Disconnect(int mode)
{
    assert(xl_running); assert(xl_app.pdispVal);
    if(!xl_running) return XL_ERR_NOT_RUNNING;

    if(mode == XL_CLOSE)
    {
        /* Application.ActiveWorkbook.Saved = True */
        VARIANT xl_workbook = { 0 }, arg = { 0 };
        arg.vt = VT_I4, arg.lVal = 1;
        HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_workbook, &xl_app, L"ActiveWorkbook", 0));
        HM_CHECK(AutoWrap(DISPATCH_PROPERTYPUT, NULL, &xl_workbook, L"Saved", 1, arg));
        HM_CHECK(VariantClear(&xl_workbook)); HM_CHECK(VariantClear(&arg));

        /* Application.Quit */ 
        HM_CHECK(AutoWrap(DISPATCH_METHOD, NULL, &xl_app, L"Quit", 0));
    }

    /* Set Application = Nothing */
    HM_CHECK(VariantClear(&xl_app));
    OleUninitialize();
    xl_running = 0;
    return XL_OK;
}

/* ------------------------------------------------------------------------ */
/* Innerhalb der aktuellen Excel-Arbeitsmappe wird das Arbeitsblatt mit     */
/* dem angegebenen Namen (zum Beispiel "Tabelle1") gesucht und aktiviert.   */
/* Rueckgabewert: XL_OK, XL_ERR_NOT_RUNNING oder XL_ERR_NOT_FOUND           */
/* ------------------------------------------------------------------------ */
int XL_SelectWorksheet(const char *name)
{
    VARIANT xl_sheet = { 0 }, arg = { 0 };
    HRESULT hr;
    wchar_t wname[1024];

    assert(xl_running); assert(xl_app.pdispVal);
    if(!xl_running) return XL_ERR_NOT_RUNNING;

    /* Application.Sheets(name).Select  */
    mbstowcs(wname, name, ARR_SIZE(wname)); wname[ARR_SIZE(wname) - 1] = 0;
    arg.vt = VT_BSTR, arg.bstrVal = SysAllocString(wname); assert(arg.bstrVal);
    hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_sheet, &xl_app, L"Sheets", 1, arg);
    if(SUCCEEDED(hr)) hr = AutoWrap(DISPATCH_METHOD, NULL, &xl_sheet, L"Select", 0);

    if(FAILED(hr))
    {
        char msg[1024];
        sprintf(msg, "Failed to select worksheet \"%s\"", name);
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
    }
    
    HM_CHECK(VariantClear(&xl_sheet)); HM_CHECK(VariantClear(&arg));
    return SUCCEEDED(hr) ? XL_OK : XL_ERR_NOT_FOUND;
}

/* ------------------------------------------------------------------------ */
/* Eine Integer-Zahl wird an der angegebenen Position in das Arbeitsblatt   */
/* geschrieben. Achtung: Die Zeilen-/Spaltennummerierung ist eins-basiert,  */
/* die oberste linke Zelle befindet sich an der Position row = 1, col = 1.  */
/* Rueckgabewert: XL_OK, XL_ERR_NOT_RUNNING oder XL_ERR_WRITE_FAILED        */
/* ------------------------------------------------------------------------ */
int XL_SetInteger(int row, int col, int value)
{
    VARIANT xl_sheet = { 0 }, xl_cell = { 0 }, vValue = { 0 }, arg[2];
    HRESULT hr;

    assert(xl_running); assert(xl_app.pdispVal);
    if(!xl_running) return XL_ERR_NOT_RUNNING;

    /* With Application.ActiveSheet */
    HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_sheet, &xl_app, L"ActiveSheet", 0));

    /* .Cells(row, col).Value = 1234 */
    vValue.vt = VT_I4, vValue.lVal = value;
    arg[0].vt = VT_I4, arg[0].lVal = row;
    arg[1].vt = VT_I4, arg[1].lVal = col;
    hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_cell, &xl_sheet, L"Cells", 2, arg[1], arg[0]);
    if(SUCCEEDED(hr)) hr = AutoWrap(DISPATCH_PROPERTYPUT, NULL, &xl_cell, L"Value", 1, vValue);

    if(FAILED(hr))
    {
        char msg[1024];
        sprintf(msg, "Failed to write integer value at row = %d, column = %d", row, col);
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
    }
    HM_CHECK(VariantClear(&arg[0])); HM_CHECK(VariantClear(&arg[1]));
    HM_CHECK(VariantClear(&xl_cell)); HM_CHECK(VariantClear(&vValue));

    /* End With */
    HM_CHECK(VariantClear(&xl_sheet));
    return SUCCEEDED(hr) ? XL_OK : XL_ERR_WRITE_FAILED;
}

/* ------------------------------------------------------------------------ */
/* Eine Double-Zahl wird an der angegebenen Position in das Arbeitsblatt    */
/* geschrieben. Achtung: Die Zeilen-/Spaltennummerierung ist eins-basiert,  */
/* die oberste linke Zelle befindet sich an der Position row = 1, col = 1.  */
/* Rueckgabewert: XL_OK, XL_ERR_NOT_RUNNING oder XL_ERR_WRITE_FAILED        */
/* ------------------------------------------------------------------------ */
int XL_SetDouble(int row, int col, double value)
{
    VARIANT xl_sheet = { 0 }, xl_cell = { 0 }, vValue = { 0 }, arg[2];
    HRESULT hr;

    assert(xl_running); assert(xl_app.pdispVal);
    if(!xl_running) return XL_ERR_NOT_RUNNING;

    /* With Application.ActiveSheet */
    HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_sheet, &xl_app, L"ActiveSheet", 0));

    /* .Cells(row, col).Value = 123.456 */
    vValue.vt = VT_R8, vValue.dblVal = value;
    arg[0].vt = VT_I4, arg[0].lVal = row;
    arg[1].vt = VT_I4, arg[1].lVal = col;
    hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_cell, &xl_sheet, L"Cells", 2, arg[1], arg[0]);
    if(SUCCEEDED(hr)) hr = AutoWrap(DISPATCH_PROPERTYPUT, NULL, &xl_cell, L"Value", 1, vValue);

    if(FAILED(hr))
    {
        char msg[1024];
        sprintf(msg, "Failed to write floating point value at row = %d, column = %d", row, col);
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
    }
    HM_CHECK(VariantClear(&arg[0])); HM_CHECK(VariantClear(&arg[1]));
    HM_CHECK(VariantClear(&xl_cell)); HM_CHECK(VariantClear(&vValue));

    /* End With */
    HM_CHECK(VariantClear(&xl_sheet));
    return SUCCEEDED(hr) ? XL_OK : XL_ERR_WRITE_FAILED;
}

/* ------------------------------------------------------------------------ */
/* Ein String wird an der angegebenen Position in das Excel-Arbeitsblatt    */
/* geschrieben. Achtung: Die Zeilen-/Spaltennummerierung ist eins-basiert,  */
/* die oberste linke Zelle befindet sich an der Position row = 1, col = 1.  */
/* Rueckgabewert: XL_OK, XL_ERR_NOT_RUNNING oder XL_ERR_WRITE_FAILED        */
/* ------------------------------------------------------------------------ */
int XL_SetString(int row, int col, const char *value)
{
    VARIANT xl_sheet = { 0 }, xl_cell = { 0 }, vValue = { 0 }, arg[2];
    HRESULT hr;
    wchar_t wvalue[1024];
 
    assert(xl_running); assert(xl_app.pdispVal);
    if(!xl_running) return XL_ERR_NOT_RUNNING;

    /* With Application.ActiveSheet */
    HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_sheet, &xl_app, L"ActiveSheet", 0));

    /* .Cells(row, col).Value = "Test" */
    arg[0].vt = VT_I4, arg[0].lVal = row;
    arg[1].vt = VT_I4, arg[1].lVal = col;
    mbstowcs(wvalue, value, ARR_SIZE(wvalue)); wvalue[ARR_SIZE(wvalue) - 1] = 0;
    vValue.vt = VT_BSTR, vValue.bstrVal = SysAllocString(wvalue); assert(vValue.bstrVal);
    hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_cell, &xl_sheet, L"Cells", 2, arg[1], arg[0]);
    if(SUCCEEDED(hr)) hr = AutoWrap(DISPATCH_PROPERTYPUT, NULL, &xl_cell, L"Value", 1, vValue);

    if(FAILED(hr))
    {
        char msg[1024];
        sprintf(msg, "Failed to write text at row = %d, column = %d", row, col);
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
    }
    HM_CHECK(VariantClear(&arg[0])); HM_CHECK(VariantClear(&arg[1]));
    HM_CHECK(VariantClear(&xl_cell)); HM_CHECK(VariantClear(&vValue));

    /* End With */
    HM_CHECK(VariantClear(&xl_sheet));
    return SUCCEEDED(hr) ? XL_OK : XL_ERR_WRITE_FAILED;
}

// Variante von XL_SetString() fuer wchar_t-Zeichenketten.
int XL_SetWString(int row, int col, const wchar_t *wvalue)
{
    VARIANT xl_sheet = { 0 }, xl_cell = { 0 }, vValue = { 0 }, arg[2];
    HRESULT hr;
    //wchar_t wvalue[1024];

    assert(xl_running); assert(xl_app.pdispVal);
    if(!xl_running) return XL_ERR_NOT_RUNNING;

    /* With Application.ActiveSheet */
    HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_sheet, &xl_app, L"ActiveSheet", 0));

    /* .Cells(row, col).Value = "Test" */
    arg[0].vt = VT_I4, arg[0].lVal = row;
    arg[1].vt = VT_I4, arg[1].lVal = col;
    //mbstowcs(wvalue, value, ARR_SIZE(wvalue)); wvalue[ARR_SIZE(wvalue) - 1] = 0;
    vValue.vt = VT_BSTR, vValue.bstrVal = SysAllocString(wvalue); assert(vValue.bstrVal);
    hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_cell, &xl_sheet, L"Cells", 2, arg[1], arg[0]);
    if(SUCCEEDED(hr)) hr = AutoWrap(DISPATCH_PROPERTYPUT, NULL, &xl_cell, L"Value", 1, vValue);

    if(FAILED(hr))
    {
        char msg[1024];
        sprintf(msg, "Failed to write text at row = %d, column = %d", row, col);
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
    }
    HM_CHECK(VariantClear(&arg[0])); HM_CHECK(VariantClear(&arg[1]));
    HM_CHECK(VariantClear(&xl_cell)); HM_CHECK(VariantClear(&vValue));

    /* End With */
    HM_CHECK(VariantClear(&xl_sheet));
    return SUCCEEDED(hr) ? XL_OK : XL_ERR_WRITE_FAILED;
}

/* ------------------------------------------------------------------------ */
/* Die Integer-Zahl an der angegebenen Position des Arbeitsblatts wird      */
/* zurueckgegeben. Hinweis: Falls die Zelle an der angegebenen Position     */
/* leer ist, keine Zahl enthaelt oder ein Fehler auftritt (falls z. B.      */
/* Excel nicht gestartet wurde), ist der Rckgabewert gleich null.          */
/* ------------------------------------------------------------------------ */
int XL_GetInteger(int row, int col)
{
    VARIANT xl_sheet = { 0 }, xl_cell = { 0 }, tmp = { 0 }, arg[2];
    HRESULT hr;
    int result = 0;

    assert(xl_running); assert(xl_app.pdispVal);
    if(!xl_running) return 0;

    /* With Application.ActiveSheet */
    HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_sheet, &xl_app, L"ActiveSheet", 0));

    /* result = CInt( .Cells(row, col).Value ) */
    VariantInit(&tmp);
    arg[0].vt = VT_I4, arg[0].lVal = row;
    arg[1].vt = VT_I4, arg[1].lVal = col;
    hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_cell, &xl_sheet, L"Cells", 2, arg[1], arg[0]);
    if(SUCCEEDED(hr)) hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &tmp, &xl_cell, L"Value", 0);
    if(SUCCEEDED(hr)) hr = VariantChangeType(&tmp, &tmp, 0, VT_I4);
    if(SUCCEEDED(hr)) result = tmp.intVal;
    else
    {
        char msg[1024];
        sprintf(msg, "Failed to read integer value at row = %d, column = %d", row, col);
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
    }

    HM_CHECK(VariantClear(&arg[0])); HM_CHECK(VariantClear(&arg[1]));
    HM_CHECK(VariantClear(&xl_cell)); HM_CHECK(VariantClear(&tmp));

    /* End With */
    HM_CHECK(VariantClear(&xl_sheet));
    return result;
}

/* ------------------------------------------------------------------------ */
/* Die Double-Zahl an der angegebenen Position des Arbeitsblatts wird       */
/* zurueckgegeben. Hinweis: Falls die Zelle an der angegebenen Position     */
/* leer ist, keine Zahl enthaelt oder ein Fehler auftritt (falls z. B.      */
/* Excel nicht gestartet wurde), ist der Rckgabewert gleich null.          */
/* ------------------------------------------------------------------------ */
double XL_GetDouble(int row, int col)
{
    HRESULT hr;
    VARIANT xl_sheet = { 0 }, xl_cell = { 0 }, tmp = { 0 }, arg[2];
    double result = 0;

    assert(xl_running); assert(xl_app.pdispVal);
    if(!xl_running) return 0;

    /* With Application.ActiveSheet */
    HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_sheet, &xl_app, L"ActiveSheet", 0));

    /* result = CDbl( .Cells(row, col).Value ) */
    VariantInit(&tmp);
    arg[0].vt = VT_I4, arg[0].lVal = row;
    arg[1].vt = VT_I4, arg[1].lVal = col;
    hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_cell, &xl_sheet, L"Cells", 2, arg[1], arg[0]);
    if(SUCCEEDED(hr)) hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &tmp, &xl_cell, L"Value", 0);
    if(SUCCEEDED(hr)) hr = VariantChangeType(&tmp, &tmp, 0, VT_R8);
    if(SUCCEEDED(hr)) result = tmp.dblVal;
    else
    {
        char msg[1024];
        sprintf(msg, "Failed to read floating point value at row = %d, column = %d", row, col);
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
    }

    HM_CHECK(VariantClear(&arg[0])); HM_CHECK(VariantClear(&arg[1]));
    HM_CHECK(VariantClear(&xl_cell)); HM_CHECK(VariantClear(&tmp));

    /* End With */
    HM_CHECK(VariantClear(&xl_sheet));
    return result;
}

/* ------------------------------------------------------------------------ */
/* Der Zellen-Inhalt an der angegebenen Position des Arbeitsblatts wird     */
/* als String zurueckgegeben. Beim Aufruf der Funktion XL_GetString muss    */
/* ein Zeiger (result) auf einen Buffer von ausreichender Groesse (len)     */
/* fuer die Rueckgabe des Zellen-Inhalts bereitgestellt werden.             */
/* Rueckgabewert: XL_OK, XL_ERR_NOT_RUNNING oder XL_ERR_READ_FAILED         */
/* ------------------------------------------------------------------------ */
int XL_GetString(int row, int col, char *result, size_t len)
{
    HRESULT hr;
    VARIANT xl_sheet = { 0 }, xl_cell = { 0 }, tmp = { 0 }, arg[2];

    assert(xl_running); assert(xl_app.pdispVal);
    if(!xl_running) return XL_ERR_NOT_RUNNING;

    assert(len > 0);
    if(len < 1) return XL_ERR_READ_FAILED;
    result[0] = 0;

    /* With Application.ActiveSheet */
    HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_sheet, &xl_app, L"ActiveSheet", 0));

    /* result = CStr( .Cells(row, col).Value ) */
    VariantInit(&tmp);
    arg[0].vt = VT_I4, arg[0].lVal = row;
    arg[1].vt = VT_I4, arg[1].lVal = col;
    hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_cell, &xl_sheet, L"Cells", 2, arg[1], arg[0]);
    if(SUCCEEDED(hr)) hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &tmp, &xl_cell, L"Value", 0);
    if(SUCCEEDED(hr)) hr = VariantChangeType(&tmp, &tmp, 0, VT_BSTR);
    if(SUCCEEDED(hr)) wcstombs(result, tmp.bstrVal, len), result[len - 1] = 0;
    else
    {
        char msg[1024];
        sprintf(msg, "Failed to read text at row = %d, column = %d", row, col);
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
    }
    HM_CHECK(VariantClear(&arg[0])); HM_CHECK(VariantClear(&arg[1]));
    HM_CHECK(VariantClear(&xl_cell)); HM_CHECK(VariantClear(&tmp));

    /* End With */
    HM_CHECK(VariantClear(&xl_sheet));
    return SUCCEEDED(hr) ? XL_OK : XL_ERR_READ_FAILED;
}

/* ----------------------------------------------------------------------- */
/* Eine Zelle (Reihe = row, Spalte = col) im aktuellen Excel-Arbeitsblatt  */
/* wird darauf ueberprueft, ob sie leer ist oder nicht. Rueckgabewert:     */
/* XL_EMPTY, XL_NOT_EMPTY, XL_ERR_READ_FAILED oder XL_ERR_NOT_RUNNING      */
/* ----------------------------------------------------------------------- */
int XL_IsEmpty(int row, int col)
{
    int result;
    HRESULT hr;
    VARIANT xl_sheet = { 0 }, xl_cell = { 0 }, tmp = { 0 }, arg[2];

    assert(xl_running); assert(xl_app.pdispVal);
    if(!xl_running) return XL_ERR_NOT_RUNNING;
    result = XL_ERR_READ_FAILED;

    /* With Application.ActiveSheet */
    HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_sheet, &xl_app, L"ActiveSheet", 0));

    /* result = CDbl( .Cells(row, col).Value ) */
    VariantInit(&tmp);
    arg[0].vt = VT_I4, arg[0].lVal = row;
    arg[1].vt = VT_I4, arg[1].lVal = col;
    hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_cell, &xl_sheet, L"Cells", 2, arg[1], arg[0]);
    if(SUCCEEDED(hr)) hr = AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &tmp, &xl_cell, L"Value", 0);
    if(SUCCEEDED(hr)) result = (tmp.vt == VT_EMPTY ? XL_EMPTY : XL_NOT_EMPTY);
    else
    {
        char msg[1024];
        sprintf(msg, "Failed to check cell at row = %d, column = %d", row, col);
        MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
    }
    HM_CHECK(VariantClear(&arg[0])); HM_CHECK(VariantClear(&arg[1]));
    HM_CHECK(VariantClear(&xl_cell)); HM_CHECK(VariantClear(&tmp));

    /* End With */
    HM_CHECK(VariantClear(&xl_sheet));
    return result;
}

/* ----------------------------------------------------------------------- */
/* Die aktuelle Arbeitsmappe wird unter dem angegebenen Dateinamen auf der */
/* Festplatte gespeichert. Rckgabewerte: XL_OK, XL_ERR_NOT_RUNNING oder   */
/* XL_ERR_SAVE_FAILED (falls beim Speichern ein Fehler aufgetreten ist).   */
/* ----------------------------------------------------------------------- */
int XL_SaveAs(const char *filename)
{
    char msg[1024];
    wchar_t wname[1024];
    HRESULT hr;
    VARIANT xl_workbook = { 0 }, arg = { 0 };

    assert(xl_running); assert(xl_app.pdispVal);
    if(!xl_running) return XL_ERR_NOT_RUNNING;

    /* Application.ActiveWorkbook.SaveAs(filename) */
    mbstowcs(wname, filename, ARR_SIZE(wname)); wname[ARR_SIZE(wname) - 1] = 0;
    arg.vt = VT_BSTR; arg.bstrVal = SysAllocString(wname); assert(arg.bstrVal);

    HM_CHECK(AutoWrap(DISPATCH_PROPERTYGET | DISPATCH_METHOD, &xl_workbook, &xl_app, L"ActiveWorkbook", 0));
    hr = AutoWrap(DISPATCH_METHOD, NULL, &xl_workbook, L"SaveAs", 1, arg);
    HM_CHECK(VariantClear(&arg)); HM_CHECK(VariantClear(&xl_workbook));
    if(SUCCEEDED(hr)) return XL_OK;

    /*  Fehler beim Speichern: Fehlermeldung ausgeben! */
    sprintf(msg, "Failed to save file: \"%s\"", filename);
    MessageBoxA(NULL, msg, "Excel Automation Error", MB_ICONEXCLAMATION);
    return XL_ERR_SAVE_FAILED;
}
